#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/uaccess.h>
#include <linux/unistd.h> 

asmlinkage long (*old_sys_write) (
       unsigned int fd, const char __user *buf, size_t count );

asmlinkage long new_sys_write (
       unsigned int fd, const char __user *buf, size_t count ) {
   int n;
   if( 1 == fd ) {
      static const char prefix[] = ":-) ";
      mm_segment_t fs = get_fs();
      set_fs( get_ds() );
      n = old_sys_write( 1, prefix, strlen( prefix ) );
      set_fs(fs);
   } 
   n = old_sys_write( fd, buf, count );
   return n;
};
EXPORT_SYMBOL( new_sys_write );

static void* find_sym( const char *sym ) {
   static unsigned long faddr = 0; // static!!!
   // ----------- вложенная функция - расширение GCC ---------
   int symb_fn( void* data, const char* sym, struct module* mod, unsigned long addr ) {
      if( 0 == strcmp( (char*)data, sym ) ) {
         faddr = addr;
         return 1;
      }
      else return 0;
   };
   // --------------------------------------------------------
   kallsyms_on_each_symbol( symb_fn, (void*)sym );
   return (void*)faddr;
}

static void **taddr; // адрес таблицы sys_call_table 

// 16 бит WP:          |
//                     V
//   3    2    2    1    1    1    0    0  0
//   1    7    3    9    5    1    7    3  0 
//   0000 0000 0000 0001 0000 0000 0000 0000 => 0x00010000
//   1111 1111 1111 1110 1111 1111 1111 1111 => 0xfffeffff


// показать управляющий регистр CR0
#define show_cr0()                        \
{  register unsigned r_eax asm ( "eax" ); \
   asm( "pushl %eax" );                   \
   asm( "movl %cr0, %eax" );              \
   printk( "! CR0 = %x\n", r_eax );       \
   asm( "popl %eax");                     \
}

//код выключения защиты записи:
#define rw_enable()              \
asm( "pushl %eax \n"             \
     "movl %cr0, %eax \n"        \
     "andl $0xfffeffff, %eax \n" \
     "movl %eax, %cr0 \n"        \
     "popl %eax" );

//код включения защиты записи:
#define rw_disable()             \
asm( "pushl %eax \n"             \
     "movl %cr0, %eax \n"        \
     "orl $0x00010000, %eax \n"  \
     "movl %eax, %cr0 \n"        \
     "popl %eax" );

static int __init wrchg_init( void ) {
   void *waddr; 
   if( ( taddr = find_sym( "sys_call_table" ) ) != NULL )
      printk( "! адрес sys_call_table = %p\n", taddr );
   else {
      printk( "! sys_call_table не найден\n" );
      return -EINVAL;
   }
   old_sys_write = (void*)taddr[ __NR_write ];
   printk( "! адрес в позиции %d[__NR_write] = %p\n", __NR_write, old_sys_write );
   if( ( waddr = find_sym( "sys_write" ) ) != NULL ) 
      printk( "! адрес sys_write = %p\n", waddr );
   else {
      printk( "! sys_call_table не найден\n" );
      return -EINVAL;
   }
   if( old_sys_write != waddr ) {
      printk( "! непонятно! : адреса не совпадают\n" );
      return -EINVAL;
   }
   printk( "! адрес нового sys_write = %p\n", &new_sys_write );
   show_cr0();
   rw_enable();
   taddr[ __NR_write ] = new_sys_write;
   show_cr0();
   rw_disable();
   show_cr0();
   return 0;
}

static void __exit wrchg_exit( void ) {
   printk( "! адрес sys_write при выгрузке = %p\n", (void*)taddr[ __NR_write ] );
   rw_enable();
   taddr[ __NR_write ] = old_sys_write;
   rw_disable();
   return;
}

module_init( wrchg_init );
module_exit( wrchg_exit );

MODULE_LICENSE( "GPL" );
MODULE_AUTHOR( "Oleg Tsiliuric <olej@front.ru>" );

/* для 64-бит
//код выключения защиты:
asm("pushq %rax");
asm("movq %cr0, %rax");
asm("andq $0xfffffffffffeffff, %rax");
asm("movq %rax, %cr0");
asm("popq %rax");

//код включения защиты:
asm("pushq %rax");
asm("movq %cr0, %rax");
asm("xorq $0x0000000000001000, %rax");
asm("movq %rax, %cr0");
asm("popq %rax");
*/
